from .version import __version__

citation = "Wang, W., Zheng, S., Shin, C. S. & Yuan, G. C. Characterizing Spatially Continuous Variations in Tissue Microenvironment through Niche Trajectory Analysis. https://www.biorxiv.org/content/10.1101/2024.04.23.590827v1. bioRxiv, 2024."